# encoding=CP1251
import nltk, os, math
import cPickle
from sys import stdout
import pymorphy


numberUnigrams = 4
numberBigrams = 4
numberTrigrams = 1
pathToCollection = "./Data/Keywords_train_data_rus/Data/"
pathToModel = "./models/ngram-ubt.pkl"

class MyRegTagger():
    def __init__(self):
        self.regexp_tagger = nltk.RegexpTagger([(r'^-?[0-9]+(.[0-9]+)?$', 'CD'),   # цифра            
            (r'более|менее|очень|крайне|скоре|некотор.*|кажд|други|котор.*|когда|однако|если|чтобы|хотя|смотря|как|также|так|зато|что|или|потом|эт|тог|тоже|словно|ежели|кабы|коли|ничем|чем', 'IN'),   # союз
            (r'ивш$|.*ывш$|.*ующ$|.*ущи$|.*ющи$|.*ящий$|.*щих$|.*щие$|.*ляя$','PRT'), # причастие
            (r'.*ими$|.*ыми$|.*ей$|.*ий$|.*ый$|.*ой$|.*его$|.*ого$|.*ую$|.*юю$|.*ая$|.*яя$|.*ою$|.*ею$','ADJ'), # прилагательное
            (r'.*чно$|.*еко$|.*соко$|.*боко$|.*роко$|.*имо$|.*мно$|.*жно$|.*жко$|.*ело$|.*тно$|.*льно$|.*здо$|.*зко$','ADV'), # наречие
            (r'.*', 'NN')])  # сущиствительное
    
    def wordTag(self,word):
        return self.regexp_tagger.tag([word.encode('utf-8')])[0][1]


class Tagger():
    def __init__(self):
        self.__tagger = pymorphy.get_morph('./pymorphy/dicts/rus')

    def wordTag(self,word):
        res = self.__tagger.get_graminfo(word.upper())
        return res[0]['class'] if res else 'NN'



# all methods for text processing
class TextProcessor():
    def __init__(self, text):
        self.tagger = MyRegTagger()
        text = text[:len(text)/4]
        self.sents = text.split(u'.')
        self.stemmer = nltk.stem.snowball.RussianStemmer()
        self.sents = [filter(lambda x: len(x) > 3,nltk.PunktWordTokenizer().tokenize(s)) for s in self.sents]
        words = sum(self.sents,[])
        self.sents = [[self.stemmer.stem(w) for w in s] for s in self.sents]
        self.words = sum(self.sents,[])
        self.originalForm = self.saveOriginalFormOfWords(words)
        self.tagged_words = dict([(w,self.tagger.wordTag(self.wordRestorer(w))) for w in set(self.words)])
    
    # save original form of word
    # here it's first occurred word of the text corresponded to the particular token
    def saveOriginalFormOfWords(self, tokenizedText):            
        originalForm = dict()
        for i in xrange(len(self.words)):
            if not(originalForm.has_key(self.words[i])):
                originalForm[self.words[i]] = tokenizedText[i]
        return originalForm
    
    # simple word restorer
    def wordRestorer(self, word):
        return self.originalForm[word]

        
# training of algorithms
# here we only compute IDf
class Trainer():
    # model is a pair of (dict: word -> idf_score) and (number of files in training set)
    model = None
    
    def __init__ (self):
        if not(self.loadModel()):
            self.model = self.computeIDF()
            self.saveModel()
        
    def computeIDF(self):
        fileCounter = 0
        docs = dict()
        for fileName in os.listdir(pathToCollection):
            fileCounter += 1
                        
            f = open(pathToCollection + fileName)
            text = unicode(f.read(), "utf-8")
            f.close()
            
            # preprocess text
            tp = TextProcessor(text) 
            
            # process unigrams
            for unigram in set(tp.words):
                docs[unigram] = docs.setdefault(unigram, 0.0) + 1.0
            
            # process bigrams
            for bigram in set(NgramProcessor().getNgrams(2,tp)):
                docs[bigram] = docs.setdefault(bigram, 0.0) + 1.0
            
            # process trigrams
            for trigram in set(NgramProcessor().getNgrams(3,tp)):
                docs[trigram] = docs.setdefault(trigram, 0.0) + 1.0
            
        # compute IDF
        idf = dict()
        for ngram in docs.keys():
            if docs[ngram] > 1:
                idf[ngram] = -math.log(docs[ngram] / fileCounter)    
        return (idf, fileCounter)
    
    def saveModel(self):
        f = open(pathToModel, "wb")
        cPickle.dump(self.model, f)
        f.close()
    
    def loadModel(self):
        try:
            f = open(pathToModel)
            self.model = cPickle.load(f)
            f.close()
            return True
        except:
            return False
        
    def getModel(self):
        return self.model


# all functions for processing n-gramms
class NgramProcessor():
    def getNgrams(self, n, tp):
        ngramList = []
        if n == 1:
            ngramList = [w for w in tp.words if tp.tagged_words[w] == 'NN']
            # ngramList = [w for w in ngramList if self.tagger.wordTag(tp.wordRestorer(w)) == u'\u0421']
            return ngramList
        for s in tp.sents:
            ngramList += zip(*([s[:-n+1]]+[s[i:-n+1+i] for i in range(n)[1:-1]]+[s[n-1:]]))
        ngramList = [ngram for ngram in ngramList if tp.tagged_words[ngram[-1]] == 'NN']
        # ngramList = [ngram for ngram in ngramList if self.tagger.wordTag(tp.wordRestorer(ngram[-1])) == u'\u0421']
        return ngramList
    
    # return sorted list of pairs (object, number of its occurrences in the list)
    # it is used for counting unigrams and bigrams in text
    def objectCounter(self, listOfObjects):
        return [(obj, 1.0*listOfObjects.count(obj)) for obj in set(listOfObjects)] 
        
    # return TopN object (unigram of bigram) from list    
    def getTopN(self, listOfObjectsWithCounts, N):
        sortedListOfObjectsWithCounts = sorted(listOfObjectsWithCounts, key=lambda x:-int(x[1]))[:N]
        # return sortedListOfObjectsWithCounts
        return [item[0] for item in sortedListOfObjectsWithCounts]

    # Assign weights to unigrams and bigrams
    # here we use TF*IDF weighting scheme
    def weightingScheme(self, listOfCountedObjects, model):
        return [(objectName, objectCount * model[0].setdefault(objectName, -math.log(1.0 / model[1]))) 
                for objectName, objectCount in listOfCountedObjects]
    
    def tStudent(self,listOfCountedObjects,listOfCountedUnigrams,model):
        resultList = []
        counterUnigrams = dict(listOfCountedUnigrams)
        for objName, objCount in listOfCountedObjects:
            
            xAverage = objCount*model[0].setdefault(objName, -math.log(1.0 / model[1]))
            nullHypothesis = 1.0
            for objUnigram in objName: nullHypothesis *= counterUnigrams[objUnigram]*(model[0].setdefault(objUnigram, -math.log(1.0 / model[1])))
            resultList.append((objName,math.sqrt(nullHypothesis)*xAverage))
        return resultList


class Keywords():
    
    def __init__(self):
        # load trained modeln
        self.model = Trainer().getModel()
        
    def getKeywords(self, text):
        # text processor
        tp = TextProcessor(text)
        #return tp
        ng = NgramProcessor()
        unigrams = ng.getNgrams(1,tp)
        bigrams = ng.getNgrams(2,tp)
        trigrams = ng.getNgrams(3,tp)
        wordsStatistic = ng.weightingScheme(ng.objectCounter(tp.words), self.model)
        unigramKeywords = map(tp.wordRestorer, #restore original forms
                              ng.getTopN(ng.weightingScheme(ng.objectCounter(unigrams), self.model), numberUnigrams))
        
        bigramKeywords = [tp.wordRestorer(word[0]) + " " + tp.wordRestorer(word[1]) 
                          for word in ng.getTopN(ng.weightingScheme(ng.objectCounter(bigrams),
                                                                  self.model), numberBigrams)]
        trigramKeywords = [' '.join([tp.wordRestorer(w) for w in word])
                          for word in ng.getTopN(ng.weightingScheme(ng.objectCounter(trigrams),
                                                                  self.model), numberTrigrams)]
        """
        T-Studen
        bigramKeywords = [tp.wordRestorer(word[0]) + " " + tp.wordRestorer(word[1])
                          for word in ng.getTopN(ng.tStudent(ng.objectCounter(bigrams),wordsStatistic,
                                                                  self.model), numberBigrams)]
        trigramKeywords = [' '.join([tp.wordRestorer(w) for w in word])
                            for word in ng.getTopN(ng.tStudent(ng.objectCounter(trigrams),wordsStatistic,
                                                                  self.model), numberTrigrams)]
        """
        return unigramKeywords + bigramKeywords + trigramKeywords

"""
def stemNgramm(ngram):
	return ' '.join([nltk.stem.snowball.RussianStemmer().stem(w) for w in ngram.split(' ')])

def global_test(start = 0,filesAmount = 0):
    print 'Test with u=%d,b=%d,t=%d and ta=%.2f' % (numberUnigrams,numberBigrams,numberTrigrams,textPart)
    path2data = './Data/Keywords_train_data_rus/Data/'
    path2ans = './Data/Keywords_train_data_rus/Answers/'
    if not filesAmount: filesAmount = len(os.listdir(path2data))
    stdout.write("\rCount: 0%")
    stdout.flush()
    k = Keywords()
    relAmount = retAmount = trueRelAmount = 0
    for i,path2file in enumerate(os.listdir(path2data)):
    	stdout.write("\rCount: %d/%d\t(%d,%d,%d)" % (i,filesAmount,trueRelAmount,relAmount,retAmount))
    	stdout.flush()
    	f = open(path2data + path2file,'r')
    	text = unicode(f.read(),'utf-8')
    	f.close()
    	res = map(stemNgramm,k.getKeywords(text))
    	try:
    		f = open(path2ans + path2file,'r')
    		ans = map(stemNgramm,unicode(f.read(),'utf-8').split('\n'))
    		f.close()
    	except:
    		continue
    	relAmount += len(res)
    	retAmount += len(ans)
    	trueRelAmount += len(set(res).intersection(set(ans)))
    recall = float(trueRelAmount)/retAmount
    precision = float(trueRelAmount)/relAmount
    fmesure = 2*recall*precision/(recall+precision)  if recall or precision else 0
    print "\nRecall = %.4f\tPrecision = %.4f\tf-mesure = %.4f" % (recall,precision,fmesure)
    return relAmount,retAmount,trueRelAmount

## -=test=-
def test():
	kw = Keywords()
	f = open("test.txt")
	text = unicode(f.read(), "utf-8")
	f.close()
	return kw.getKeywords(text)
#print repr(keywords).decode("unicode_escape").encode("cp1251")
"""